# Because we are compatible with PowerShell 5, standard backtick-escaping doesn't work here.
# Therefore, escape sequences are often cast using [char] or a static method thereof.

$Esc = [char]27 # `e
$Bold = "$($Esc)[1m"
$Reset = "$($Esc)[0m"

function __gk_8_bit_color([string]$Color) {
  return "$($Esc)[38;5;$($Color)m"
}

function __gk_apply_style([string]$Style, [string]$Text) {

  # A note about icons:
  # This file will not be interpreted properly as UTF-8 by PowerShell without a very specific BOM, which most editors
  # don't support because it's wrong. Therefore we really need to be careful to stay inside ASCII for minimum issues.

  switch($Style) {
    "ahead" {
      $Style = "$($Bold)$(__gk_8_bit_color("10"))"
      $Icon = "$([char]::ConvertFromUtf32(0xA71B)) " # ꜛ
    }
    "behind" {
      $Style = "$($Bold)$(__gk_8_bit_color("208"))"
      $Icon = "$([char]::ConvertFromUtf32(0x2193)) " # ↓
    }
    "bold" {
      $Style = "$($Bold)"
      $Icon = ""
    }
    "branch" {
      $Style = "$($Bold)$(__gk_8_bit_color("5"))"
      $Icon = "$([char]::ConvertFromUtf32(0x1F331)) " # 🌱
    }
    "repository" {
      $Style = "$($Bold)$(__gk_8_bit_color("6"))"
      $Icon = ""
    }
    "tag" {
      $Style = "$($Bold)$(__gk_8_bit_color("196"))"
      $Icon = "$([char]::ConvertFromUtf32(0x1F3F7)) " # 🏷
    }
    default {
      $Style = ""
      $Icon = ""
    }
  }

  return "$($Reset)$($Style)$($Icon)$($Text)$($Reset)"
}

function __gk_get_repository_name {
  $RepoPath = git rev-parse --show-toplevel 2>$null
  if ($RepoPath -eq $null) {
    return ""
  }

  $RepoItem = (Get-Item $RepoPath) 2>$null
  if ($RepoItem -eq $null) {
    return ""
  }

  $RepoName = $RepoItem.basename
  $NewPrompt = "$(__gk_apply_style -Style "repository" -Text "$($RepoName)")"

  $BranchInfo = git branch -v | Select-String -Pattern "\*" | Select-Object -ExpandProperty Line
  $GitRebaseNeedle = '(no branch, rebasing'
  if ($BranchInfo.Contains($GitRebaseNeedle)) {
    $BranchName = 'rebase in progress: ' + $BranchInfo.Split(' ')[4]
    $BranchName = $BranchName.trim('()')
  }
  else {
    $BranchName = $BranchInfo.Split(' ')[1]
  }
  $AheadCommitsMatch = $BranchInfo -match "^\*.*\[ahead (\d+)(\]|,)"
  $BehindCommitsMatch = $BranchInfo -match "^\*.*(?:\[|, )behind (\d+)\]"

  $TagName = git for-each-ref --points-at=HEAD --count=1 --sort=-version:refname --format='%(refname:short)' refs/tags 2>$null

  if ($BranchName) {
    if ($BranchName -eq "(HEAD") {
      $BranchName = ' DETACHED HEAD'
    }

    $NewPrompt = "$NewPrompt $(__gk_apply_style -Style "bold" -Text "on") $(__gk_apply_style -Style "branch" -Text "$($BranchName)")"
  }

  if ($AheadCommitsMatch) {
    $AheadCommits = $BranchInfo -replace '.*ahead ([0-9]+).*', '$1'
    $NewPrompt = "$NewPrompt $(__gk_apply_style -Style "ahead" -Text $AheadCommits)"
  }

  if ($BehindCommitsMatch) {
    $BehindCommits = $BranchInfo -replace '.*behind ([0-9]+).*', '$1'
    $NewPrompt = "$NewPrompt $(__gk_apply_style -Style "behind" -Text $BehindCommits)"
  }

  if ($TagName) {
    $NewPrompt = "$NewPrompt $(__gk_apply_style -Style "bold" -Text "is") $(__gk_apply_style -Style "tag" -Text $TagName)"
  }

  return "$NewPrompt`n"
}

$OriginalPrompt = $(Get-Command Prompt 2>$null).ScriptBlock

$PipeClientStream = New-Object System.IO.Pipes.NamedPipeClientStream('.', $Env:GK_CWD_PIPE, [System.IO.Pipes.PipeDirection]::Out)

function __gk_connect_pipe() {
  $PipeClientStream.Connect(5000)
  $script:PipeWriter = New-Object System.IO.StreamWriter($PipeClientStream);
  $script:PipeWriter.AutoFlush = $True
}

gk showWelcomeMessage
__gk_connect_pipe 2>$null

function Prompt {
  if ($script:PipeWriter -ne $null) {
    $cwd = Get-Location
    $script:PipeWriter.writeLine($cwd)
  }

  $p = "$(__gk_get_repository_name)$(Invoke-Command -ScriptBlock $OriginalPrompt)"
  $pLines = $p.split("`n");
  $LastLength = $pLines[$pLines.length-1].length;
  $RowOffset = $pLines.length-1;
  $LengthMarker = "$($Esc)[1314;$($LastLength);$($RowOffset)~"
  $pLines[$pLines.length-1] += $LengthMarker;

  return $pLines -join "`n";
}

# TODO: It would be nice to restore the original prompt on unload, but I haven't found a way to do so.
#       Surely the sorcery can be done by attaching a code block to
#         $MyInvocation.MyCommand.ScriptBlock.Module.OnRemove ...but we don't know exactly what the magic spell is.

Export-ModuleMember -function prompt
